from typing import Annotated, List
from fastapi import APIRouter, Body, File, UploadFile
from agentuniverse_product.service.model.knowledge_dto import KnowledgeDTO
from agentuniverse_product.service.knowledge_service.knowledge_service import KnowledgeService
from magent_ui.utils import AsyncTask

router = APIRouter()
knowledge_router = router


@router.get("/knowledge", response_model=List[KnowledgeDTO])
async def get_knowledge():
    return await AsyncTask.to_thread(KnowledgeService.get_knowledge_list)


@router.post("/knowledge", response_model=str)
async def create_knowledge(knowledge: KnowledgeDTO):
    return await AsyncTask.to_thread(KnowledgeService.create_knowledge, knowledge)


@router.put("/knowledge/{knowledge_id}", response_model=str)
async def update_knowledge(knowledge_id: str, knowledge: KnowledgeDTO):
    knowledge.id = knowledge_id
    return await AsyncTask.to_thread(KnowledgeService.update_knowledge, knowledge)


@router.delete("/knowledge/{knowledge_id}", response_model=bool)
async def delete_knowledge(knowledge_id: str):
    return await AsyncTask.to_thread(KnowledgeService.delete_knowledge, knowledge_id)


@router.post("/knowledge/upload")
async def upload_knowledge_file(knowledge_id: Annotated[str, Body()], file: UploadFile = File(...)):
    print('here')
    return await AsyncTask.to_thread(KnowledgeService.upload_knowledge_file, knowledge_id, file)



from fastapi import APIRouter
from magent_ui.routers.agents.router import agents_router
from magent_ui.routers.knowledge.router import knowledge_router
from magent_ui.routers.llms.router import llms_router
from magent_ui.routers.sessions.router import sessions_router
from magent_ui.routers.tools.router import tools_router
from magent_ui.routers.resource.router import resource_router
from magent_ui.routers.workflow.router import workflow_router
from magent_ui.routers.common.router import common_router
from magent_ui.routers.plugins.router import plugins_router
from magent_ui.routers.test.router import test_router

api_router = APIRouter()

api_router.include_router(agents_router, prefix="/v1", tags=["agent"])
api_router.include_router(knowledge_router, prefix="/v1", tags=["knowledge"])
api_router.include_router(llms_router, prefix="/v1", tags=["llm"])
api_router.include_router(sessions_router, prefix="/v1", tags=["session"])
api_router.include_router(tools_router, prefix="/v1", tags=["tool"])
api_router.include_router(resource_router, prefix="/v1", tags=["resource"])
api_router.include_router(workflow_router, prefix="/v1", tags=["workflow"])
api_router.include_router(common_router, prefix="/v1", tags=["common"])
api_router.include_router(plugins_router, prefix="/v1", tags=["plugin"])
api_router.include_router(test_router, prefix="/v1", tags=["test"])



import asyncio
from concurrent.futures import ThreadPoolExecutor
from typing import AsyncIterable, Iterator
from multiprocessing import cpu_count


class AsyncTask():
    # 创建一个较大的线程池
    executor = ThreadPoolExecutor()

    @staticmethod
    def set_max_worker(max: int):
        AsyncTask.executor = ThreadPoolExecutor(max_workers=max)

    @staticmethod
    async def to_thread(func, *args, **kwargs):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(AsyncTask.executor, func, *args, **kwargs)


# def next_item(sync_iter: Iterator):
#     """获取同步迭代器的下一个项目，如果结束则返回 StopIteration 哨兵值"""
#     try:
#         return next(sync_iter)
#     except StopIteration:
#         return StopIteration


# async def iterator_to_async_iterable(sync_iter: Iterator) -> AsyncIterable:
#     """将同步迭代器转换为异步可迭代对象"""
#     loop = asyncio.get_running_loop()
#     while True:
#         # 在线程中运行同步的 next 操作
#         item = await asyncio.to_thread(next_item, sync_iter)
#         if item is StopIteration:
#             # 如果返回值是 StopIteration 哨兵值，终止循环
#             break
#         yield item
#         await asyncio.sleep(0)  # 允许其他异步任务运行



from locust import HttpUser, task


class ApiTestUser(HttpUser):

    # @task(1)
    # def sleep(self):
    #     self.client.get("http://localhost:8888/api/v1/test/sleep")

    # @task(10)
    # def sleep_async(self):
    #     self.client.get("http://localhost:8888/api/v1/test/sleep_async")

    # @task(10)
    # def sleep_async_1(self):
    #     self.client.get("http://localhost:8888/api/v1/test/sleep_async_1")

    @task(100)
    def get_agents(self):
        self.client.get("http://localhost:8888/api/v1/agents")

    @task(1)
    def chant_stream(self):
        self.client.post(
            "http://localhost:8888/api/v1/agents/demo_rag_agent/stream-chat", json={"agent_id": "demo_rag_agent", "session_id": "16d1acec-bd6c-4b46-bb3b-f35351fd11c1", "input": "巴菲特为什么减持苹果股票"})



# config_loader.py

import logging
import os
import importlib.util
from pathlib import Path

logger = logging.getLogger("uvicorn")

default_config = {
    'thread_worker': 20,
    'host': '0.0.0.0',
    'port': 8888,
    'base_url': None,
    'open_browser': True,
    'log_level': None,
    'root_path': '/'
}


def to_uvicorn_config(config: dict):
    uvicorn_config = {**config}
    # remove unsupported config key
    del uvicorn_config['base_url']
    del uvicorn_config['open_browser']
    del uvicorn_config['log_level']

    # root path has already taken effect
    del uvicorn_config['root_path']

    if uvicorn_config.get('thread_worker', None) is not None:
        del uvicorn_config['thread_worker']

    return uvicorn_config


def load_config_from_file(file_path):
    spec = importlib.util.spec_from_file_location("user_config", file_path)
    if spec is not None:
        config_module = importlib.util.module_from_spec(spec)
        if spec.loader is not None:
            spec.loader.exec_module(config_module)
            return {key: getattr(config_module, key) for key in dir(config_module) if not key.startswith('_')}
    return {}


def load_config_from_env(default_config, prefix='MAGENT_UI_SERVER'):
    env_config = {}
    for key, value in default_config.items():
        if isinstance(value, dict):
            env_config[key] = load_config_from_env(
                value, f"{prefix}_{key.upper()}")
        else:
            env_var = f"{prefix}_{key.upper()}"
            env_value = os.getenv(env_var)
            if env_value is not None:
                if isinstance(value, bool):
                    env_value = env_value.lower() in ['true', '1', 'yes']
                elif isinstance(value, int):
                    env_value = int(env_value)
                env_config[key] = env_value
    return env_config


def merge_dicts(default, override):
    for key, value in override.items():
        if isinstance(value, dict) and key in default:
            default[key] = merge_dicts(default[key], value)
        else:
            default[key] = value
    return default


def load_config(config=default_config, project_root_path=None):
    config = merge_dicts(default_config, config)

    # 用户目录配置文件路径
    user_config_path = os.path.expanduser('~/.magent/ui_config.py')

    # 加载用户目录配置文件
    if os.path.exists(user_config_path):
        user_config = load_config_from_file(user_config_path)
        logger.info(
            f"Load user config from {user_config_path}.")
        config = merge_dicts(config, user_config)

    # 工作目录配置文件路径
    if project_root_path is None:
        project_root_path = os.getcwd()

    # 加载项目根目录配置文件
    project_config_path = os.path.join(
        project_root_path, 'config/magent_ui_config.py')
    if os.path.exists(project_config_path):
        project_config = load_config_from_file(project_config_path)
        logger.info(
            f"Load project config from {project_config_path}.")
        config = merge_dicts(config, project_config)

    project_root_config_path = os.path.join(
        project_root_path, '.magent_ui_config.py')
    if os.path.exists(project_root_config_path):
        project_config = load_config_from_file(project_root_config_path)
        logger.info(
            f"Load project config from {project_config_path}.")
        config = merge_dicts(config, project_config)

    # 加载环境变量配置
    env_config = load_config_from_env(default_config)
    if len(env_config.keys()):
        logger.info("Load env config.")
    config = merge_dicts(config, env_config)

    return config


api_path = 'api'
static_path = 'static'
resource_path = 'resources'
app_path = 'app'


config_count = 0


class AppConfig():
    config: dict
    project_root_path: Path
    resource_dir_path: Path

    port: int
    root_path: str
    base_url: str
    open_browser: bool
    log_level: str

    full_api_path: str
    full_static_path: str
    full_resource_path: str

    api_url: str
    static_url: str
    resource_url: str
    app_url: str

    thread_worker: int

    def __init__(self):
        global config_count
        config_count += 1
        print('config:', config_count)

    def load_config(self, project_root_path: Path, **kwargs):
        config = load_config(kwargs, project_root_path)
        self.project_root_path = project_root_path
        self.resource_dir_path = project_root_path / 'app' / 'resources'
        self.config = config
        self.port = config.get('port', 8888)
        base_root_path = '/'
        root_path = config.get('root_path', base_root_path)
        self.root_path = root_path
        self.base_url = config.get('base_url', None)
        self.open_browser = config.get('open_browser', True)
        self.log_level = config.get('log_level', None)
        self.thread_worker = config.get('thread_worker', None)
        if self.base_url is None:
            self.base_url = root_path

        if not root_path.startswith('/'):
            logger.info('[magent] root_path should start with "/" ', root_path)
            root_path = f'/{root_path}'
            config['root_path'] = root_path

        self.full_api_path = os.path.join(root_path, api_path)
        self.full_static_path = os.path.join(root_path, static_path)
        self.full_resource_path = os.path.join(root_path, resource_path)

        self.api_url = os.path.join(self.base_url, api_path)
        self.static_url = os.path.join(self.base_url, static_path)
        self.resource_url = os.path.join(self.base_url, resource_path)
        self.app_url = os.path.join(self.base_url, app_path)


app_config = AppConfig()



from typing import List
from fastapi import APIRouter
from agentuniverse_product.service.plugin_service.plugin_service import PluginService
from agentuniverse_product.service.model.plugin_dto import PluginDTO
from magent_ui.utils import AsyncTask

router = APIRouter()
plugins_router = router


@router.get("/plugins", response_model=List[PluginDTO])
async def get_agents():
    return await AsyncTask.to_thread(PluginService.get_plugin_list)


@router.post("/plugins/openapi", response_model=str)
async def create_plugin_with_openapi(plugin: PluginDTO):
    return await AsyncTask.to_thread(PluginService.create_plugin_with_openapi, plugin)



from fastapi import APIRouter, Form,File, UploadFile
from magent_ui.config import app_config
import aiofiles

router = APIRouter()
resource_router = router

@router.post("/resource/uploadfile")
async def upload_resource_file(file: UploadFile = File(...), filename:str= Form(...)):
    file_location = app_config.resource_dir_path.joinpath(filename)

    # 使用 aiofiles 异步写入文件
    async with aiofiles.open(file_location, "wb") as buffer:
        while content := await file.read(1024):  # 每次读取 1024 字节
            await buffer.write(content)

    return {"filename": filename}



from datetime import datetime
import time
from typing import AsyncIterable, Iterator
import sys
import asyncio
from fastapi import APIRouter

router = APIRouter()
test_router = router


class SlowSyncIterator:
    iter_time: int = 0
    max: int
    order: int

    def __init__(self, order, loop: int = 3):
        self.max = loop
        self.order = order

    def __iter__(self):
        return self

    def __next__(self):
        time.sleep(1)  # Simulating a time-consuming operation
        self.iter_time += 1
        print('slow_iter:', self.order, '-',  self.iter_time)
        print(datetime.now())
        if self.iter_time > self.max:
            print('__next__ raise StopIteration')
            raise StopIteration
        return self.iter_time


def next_item(sync_iter: Iterator):
    try:
        return next(sync_iter)
    except StopIteration:
        return StopIteration


async def iterator_to_async_iterable(sync_iter: Iterator) -> AsyncIterable:
    loop = asyncio.get_running_loop()
    while True:
        item = await asyncio.to_thread(next_item, sync_iter)
        if item is StopIteration:
            break
        yield item
        await asyncio.sleep(0)  # Allow other tasks to run


class TestService:
    @staticmethod
    def sleep(order: int = 0, t: int = 1):
        time.sleep(t)  # 模拟一个耗时操作
        print('sleep', order, '-', t)
        print(datetime.now())
        return order

    @staticmethod
    async def async_slow_iter(order: int, loop: int = 3):
        sync_iter = SlowSyncIterator(order, loop)
        async_iter = iterator_to_async_iterable(iter(sync_iter))
        final: int
        async for item in async_iter:
            print('async', datetime.now())
            final = item
        return final

    @staticmethod
    def slow_iter(order: int, loop: int = 3):
        sync_iter = SlowSyncIterator(order, loop)
        final: int
        for item in sync_iter:
            final = item
        return final

    @staticmethod
    async def async_sleep(order: int = 0, t: int = 1) -> int:
        def run():
            return TestService.sleep(order, t)
        return await asyncio.to_thread(run)


order = 0


@router.get("/test/sleep", response_model=int)
async def test_sleep():
    global order
    order += 1
    print('sync sleep', order)
    res = TestService.sleep(order)
    print('sync sleep end', order)
    return res


@router.get("/test/sleep_async", response_model=int)
async def test_sleep_async():
    global order
    order += 1
    print('async sleep', order)
    res = await TestService.async_sleep(order)
    print('async sleep end', order)
    return res


@router.get("/test/sleep_async_1", response_model=int)
async def test_sleep_async_1():
    global order
    order += 1
    print('1 async sleep', order)
    res = await TestService.async_sleep(order)
    print('1 async sleep end', order)
    return res



from typing import List
from fastapi import APIRouter
from agentuniverse_product.service.tool_service.tool_service import ToolService
from agentuniverse_product.service.model.tool_dto import ToolDTO
from magent_ui.utils import AsyncTask

router = APIRouter()
tools_router = router


@router.get("/tools", response_model=List[ToolDTO])
async def get_agents():
    return await AsyncTask.to_thread(ToolService.get_tool_list)



from fastapi import APIRouter
from magent_ui.utils import AsyncTask
from agentuniverse_product.base.util.common_util import is_component_id_unique

router = APIRouter()
common_router = router


@router.get("/common/is_id_unique", response_model=bool)
async def is_id_unique(id: str, type: str):
    return await AsyncTask.to_thread(is_component_id_unique, id, type)



from typing import List
from fastapi import APIRouter
from agentuniverse_product.service.llm_service.llm_service import LLMService
from agentuniverse_product.service.model.llm_dto import LlmDTO
from magent_ui.utils import AsyncTask

router = APIRouter()
llms_router = router


@router.get("/llms", response_model=List[LlmDTO])
async def get_agents():
    return await AsyncTask.to_thread(LLMService.get_llm_list)



from typing import List
from fastapi import APIRouter
from agentuniverse_product.service.session_service.session_service import SessionService
from agentuniverse_product.service.model.session_dto import SessionDTO
from pydantic import BaseModel
from magent_ui.utils import AsyncTask

router = APIRouter()
sessions_router = router


@router.get("/sessions", response_model=List[SessionDTO])
async def get_agents(agent_id: str):
    return await AsyncTask.to_thread(SessionService.get_session_list, agent_id)


@router.get("/sessions/{session_id}", response_model=SessionDTO | None)
async def get_session_detail(session_id):
    return await AsyncTask.to_thread(SessionService.get_session_detail, session_id)


class SessionCreate(BaseModel):
    agent_id: str


@router.post("/sessions", response_model=SessionDTO)
async def create_session(model: SessionCreate):
    session_id = SessionService.create_session(model.agent_id)
    return await AsyncTask.to_thread(SessionService.get_session_detail, session_id)


@router.delete("/sessions/{session_id}", response_model=str)
async def delete_session(session_id):
    return await AsyncTask.to_thread(SessionService.delete_session, session_id)



import asyncio
import enum
import json
from typing import AsyncIterable, List
from fastapi import APIRouter
from agentuniverse_product.service.agent_service.agent_service import AgentService
from agentuniverse_product.service.workflow_service.workflow_service import WorkflowService
from agentuniverse_product.service.model.agent_dto import AgentDTO
from agentuniverse_product.service.model.workflow_dto import WorkflowDTO
from agentuniverse_product.service.model.planner_dto import PlannerDTO
from pydantic import BaseModel
from sse_starlette import EventSourceResponse, ServerSentEvent


router = APIRouter()
agents_router = router


@router.get("/agents", response_model=List[AgentDTO])
async def get_agents():
    return await asyncio.to_thread(AgentService.get_agent_list)


@router.get("/agents/{agent_id}", response_model=AgentDTO | None)
async def get_agent_detail(agent_id):
    return await asyncio.to_thread(AgentService.get_agent_detail, agent_id)


@router.put("/agents/{agent_id}", response_model=AgentDTO | None)
async def update_agent(agent_id, agent: AgentDTO):
    return await asyncio.to_thread(AgentService.update_agent, agent)


@router.post("/agents", response_model=str)
async def create_agent(agent: AgentDTO):
    return await asyncio.to_thread(AgentService.create_agent, agent)


@router.post("/agents/workflow", response_model=str)
async def create_workflow_agent(agent: AgentDTO):
    def task():
        workflow_id = f"{agent.id}_workflow"
        workflow_name = f"{agent.nickname}_workflow"
        workflow = WorkflowDTO(id=workflow_id, name=workflow_name)
        workflow_id = WorkflowService.create_workflow(workflow)
        agent.planner = PlannerDTO(
            id='workflow_planner', workflow_id=workflow_id)
        return AgentService.create_agent(agent)
    return await asyncio.to_thread(task)


class MessageSenderType(enum.Enum):
    AI = "ai"
    HUMAN = "human"
    SYSTEM = "system"


class MessageCreate(BaseModel):
    session_id: str
    agent_id: str
    input: str


class MessageOutput(BaseModel):
    message_id: int
    session_id: str
    response_time: float
    output: str
    start_time: str
    end_time: str


@router.post("/agents/{agent_id}/chat", response_model=MessageOutput)
async def chat(agent_id, model: MessageCreate):
    output_dict = AgentService.chat(
        model.agent_id, model.session_id, model.input)
    print(output_dict)
    return MessageOutput.model_validate(output_dict)


class SSEType(enum.Enum):
    MESSAGE = "message"
    STEPS = "steps"
    BLANK_MESSAGE = "blank_message"
    CHUNK = "chunk"
    ERROR = "error"
    EOF = "EOF"
    RESULT = "result"


async def send_message(model: MessageCreate) -> AsyncIterable[ServerSentEvent]:
    msg_iterator = AgentService.async_stream_chat(
        model.agent_id, model.session_id, model.input)
    async for msg_chunk in msg_iterator:
        type = msg_chunk.get("type", None)
        if type == "error":
            yield ServerSentEvent(event=SSEType.ERROR.value, id=model.session_id, data=json.dumps(msg_chunk, ensure_ascii=False))
        if type == "token":
            yield ServerSentEvent(event=SSEType.CHUNK.value, id=model.session_id, data=json.dumps(msg_chunk, ensure_ascii=False))
        if type == "intermediate_steps":
            yield ServerSentEvent(event=SSEType.STEPS.value, id=model.session_id, data=json.dumps(msg_chunk, ensure_ascii=False))
        if type == "final_result":
            yield ServerSentEvent(event=SSEType.RESULT.value, id=model.session_id, data=json.dumps(msg_chunk, ensure_ascii=False))


@router.post("/agents/{agent_id}/stream-chat")
async def stream_chat(agent_id, model: MessageCreate):
    print(agent_id)
    print(model)
    return EventSourceResponse(send_message(model), media_type="text/event-stream")



from fastapi import FastAPI, Request
from fastapi.responses import HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from contextlib import asynccontextmanager
import webbrowser
from magent_ui.utils import AsyncTask
import uvicorn
import logging
from magent_ui.routers.main import api_router
from magent_ui.config import to_uvicorn_config, app_config
import os
from agentuniverse.base.util.system_util import get_project_root_path
from uvicorn.config import LOGGING_CONFIG

# Use uvicorn's default logging configuration
logging.config.dictConfig(LOGGING_CONFIG)  # type: ignore

# Get the uvicorn logger
logger = logging.getLogger("uvicorn")

# 挂载 static 目录，使其可以访问静态文件
BASE_DIR = os.path.dirname(os.path.abspath(__file__))
static_dir = os.path.join(BASE_DIR, 'static')
templates_dir = os.path.join(BASE_DIR, 'templates')


templates = Jinja2Templates(directory=templates_dir)


def launch(**kwargs):
    logger.info("Current log level is")
    project_root_path = get_project_root_path()

    app_config.load_config(project_root_path=project_root_path, **kwargs)

    if app_config.log_level is not None:
        logger.setLevel(app_config.log_level)

    if app_config.thread_worker is not None:
        AsyncTask.set_max_worker(app_config.thread_worker)

    @asynccontextmanager
    async def lifespan(app: FastAPI):
        if app_config.open_browser:
            url = f"http://localhost:{app_config.port}{app_config.root_path}"
            webbrowser.open(url)
        logger.info(f"Server is running at {url}")
        yield
        logger.info('Server finished')

    app = FastAPI(lifespan=lifespan)

    # api
    app.include_router(api_router, prefix=app_config.full_api_path)

    # static
    if os.path.exists(static_dir):
        app.mount(app_config.full_static_path, StaticFiles(directory=static_dir,
                                                           html=True), name="static")
    else:
        logger.info('Can not find static directory. ', static_dir)

    # resources
    if not app_config.resource_dir_path.exists():
        logger.info('Resource directory not exist. create at',
                    app_config.resource_dir_path)
        os.makedirs(app_config.resource_dir_path)

    app.mount(app_config.full_resource_path, StaticFiles(directory=app_config.resource_dir_path,
                                                         html=True))

    # auto redirect to app url
    @app.get(app_config.root_path, response_class=HTMLResponse)
    async def to_index_page(request: Request):
        return RedirectResponse(url=f"{app_config.app_url}/")

    # html as default, app_path included
    html_root = app_config.root_path if app_config.root_path.endswith(
        '/') else f"{app_config.root_path}/"

    @app.get(html_root+"{path:path}", response_class=HTMLResponse)
    async def to_app_page(request: Request):
        page_config = {
            "baseUrl": app_config.base_url,
            "resourceUrl": app_config.resource_url,
            "apiUrl": app_config.api_url,
            "appUrl": app_config.app_url,
            "staticUrl": app_config.static_url,
        }
        return templates.TemplateResponse(
            request=request, name="index.html", context={
                "page_config": page_config, "static_url": app_config.static_url
            }
        )

    uvicorn_config = to_uvicorn_config(app_config.config)
    uvicorn.run(app, log_level='info', **uvicorn_config)



from fastapi import APIRouter
from agentuniverse_product.service.workflow_service.workflow_service import WorkflowService
from agentuniverse_product.service.model.workflow_dto import WorkflowDTO
from magent_ui.utils import AsyncTask

router = APIRouter()
workflow_router = router


@router.get("/workflows/{workflow_id}", response_model=WorkflowDTO | None)
async def get_agent_detail(workflow_id):
    return await AsyncTask.to_thread(WorkflowService.get_workflow_detail, workflow_id)


@router.put("/workflows/{workflow_id}", response_model=str | None)
async def update_agent(workflow_id, workflow: WorkflowDTO):
    return await AsyncTask.to_thread(WorkflowService.update_workflow, workflow)


@router.post("/workflows", response_model=WorkflowDTO)
async def create_agent(workflow: WorkflowDTO):
    return await AsyncTask.to_thread(WorkflowService.create_workflow, workflow)
